var _            = require('lodash'),
    fn           = require('to-function'),
    chrono       = require('chrono-node'),
    trim         = require('string-trim'),
    underLoad    = require('underload'),
    HTMLParser   = require('htmlparser2'),
    unescapeHTML = require('../../support/unescape-html-entities'),
    MailParser   = require('mailparser').MailParser;

var isReplyLine = function (line) {
  return [
    /From:\s*.+?@.+?/i,
    /<.+?@.+?>/,
    /.+?@.+?\s+wrote:$/i,
    /^.*On.*(\n)?wrote:$/i,
    /\-+original\s+message\-+\s*$/i,
    /from:\s*$/i,
    /^begin forwarded message:$/i
  ].some(function (regex) {
    return regex.test(line);
  });
};

var isForwardLine = function (line) {
  return (/^(?:from|subject|date|to):\s+[^$]/i).test(line);
};

var makeConversation = function (conversation) {
  return conversation || {
    text: ''
  };
};

// Writing my own text parsing module since I need to support line breaks and
// block/inline elements
var parseDomToText = function getText (elem) {
  // Adding br to the inline elements list since it is special cased here
  var inlineElements = ['br', 'a', 'span', 'small'],
      text = '';

  if (Array.isArray(elem)) {
    return elem.map(getText).join('');
  }
  if (elem.type === 'tag') {
    text += (elem.name === 'br' ? '\n' : '');
    text += getText(elem.children);
    text += (!~inlineElements.indexOf(elem.name) && (elem.prev && !~inlineElements.indexOf(elem.prev.name)) ? '\n' : '');
    return text;
  }
  if (elem.type === 'text') {
    return unescapeHTML(elem.data.replace(/\s+/g, ' '));
  }
  return text;
};

module.exports = function (email, cb) {
  var parser = new MailParser();
  parser.on('end', function (email) {
    var startDate     = chrono.parseDate(email.headers.date),
        timeDiff      = new Date(email.headers.date).getTime() - startDate.getTime(),
        forwardText   = '',
        converseNum   = 0,
        forwardLength = 0,
        parser,
        parseLine;

    email.conversation = [];

    parseLine = function (line) {
      var prevForwardText   = forwardText,
          prevForwardLength = forwardLength,
          conversation,
          date;

      line         = trim(trim.left(line, ' >'));
      conversation = email.conversation[converseNum] = makeConversation(email.conversation[converseNum]);

      if (isForwardLine(line)) {
        forwardText   += line + '\n';
        forwardLength += 1;
      } else {
        forwardText   = '';
        forwardLength = 0;
      }

      if (isReplyLine(line)) {
        converseNum += 1; // Increase the conversation number every time we see a reply line

        conversation = email.conversation[converseNum] = makeConversation(email.conversation[converseNum]);
        date         = chrono.parse(line, email.conversation[converseNum - 1].date).map(fn('start')).reduce(underLoad.binary(_.defaults), {});

        // Create a new date object that will hold the conversation date
        if (date && date.year && date.month && date.day) {
          conversation.date = new Date(date.year, date.month, date.day, date.hour || 0, date.minute || 0, date.second || 0);
        }
        return;
      }

      if (prevForwardLength > 3 && !forwardLength) {
        return conversation.date = chrono.parse(prevForwardText || email.headers.date, email.conversation[converseNum - 1].date).pop().startDate;
      }

      // If the email is forwarded
      if (forwardLength) { return; }

      // Append the text to the conversation
      conversation.text += line + '\n';
      // If the date couldn't be parsed from the message, use the previous
      // messages date - or the original emails send date
      conversation.date = conversation.date || (email.conversation[converseNum - 1] &&
        email.conversation[converseNum - 1].date) || startDate;
    };

    parser = new HTMLParser.createDomStream(function (_, dom) {
      parseDomToText(dom).split('\n').forEach(parseLine);

      // Run through the conversations and trim messages, removing any empty messages
      email.conversation = email.conversation.filter(function (message, index, conversation) {
        message.text = trim(message.text);
        // Apply the time difference to all the dates
        message.date && (message.date = new Date(message.date.getTime() + timeDiff));
        return message.text;
      });

      cb(email);
    });
    parser.write(email.html);
    parser.end();
  });
  parser.write(email);
  parser.end();
};
